@@ -109,6 +109,12 @@ void snd_compat_request_module(const char *name, ...);
 #ifndef minor
 #define minor(x) MINOR(x)
 #endif
+#ifndef imajor
+#define imajor(x) major((x)->i_rdev)
+#endif
+#ifndef iminor
+#define iminor(x) minor((x)->i_rdev)
+#endif
 #ifndef mk_kdev
 #define mk_kdev(maj, min) MKDEV(maj, min)
 #endif
@@ -121,7 +127,7 @@ void snd_compat_request_module(const char *name, ...);
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 3) && !defined(need_resched)
 #define need_resched() (current->need_resched)
 #endif
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 4)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 4) && !defined(WE_ARE_WOLK)
 #include <linux/fs.h>
 static inline struct proc_dir_entry *PDE(const struct inode *inode)
 {
@@ -187,18 +193,18 @@ int snd_compat_vsnprintf(char *buf, size_t size, const char *fmt, va_list args);
 #define writeq(v, a) do { __writeq((v),(a)); mb(); } while(0)
 #endif
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 28)
 #include <linux/interrupt.h>
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 28)
 static inline void synchronize_irq_wrapper(unsigned int irq) { synchronize_irq(); }
 #undef synchronize_irq
 #define synchronize_irq(irq)	synchronize_irq_wrapper(irq)
 #endif /* LINUX_VERSION_CODE < 2.5.28 */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 68)
+#ifndef IRQ_NONE
 #define IRQ_NONE	/*void*/
 #define IRQ_HANDLED	/*void*/
 #define IRQ_RETVAL(x)	/*void*/
 typedef void irqreturn_t;
-#endif /* LINUX_VERSION_CODE < 2.5.68 */
+#endif
 
 #ifndef min
 /*
@@ -314,6 +320,7 @@ int snd_hack_usb_set_interface(struct usb_device *dev, int interface, int altern
 #define usb_host_config		usb_config_descriptor
 #define usb_host_interface	usb_interface_descriptor
 #define usb_host_endpoint	usb_endpoint_descriptor
+#define get_iface(cfg, num)	(&(cfg)->interface[num])
 #define get_iface_desc(iface)	(iface)
 #define get_endpoint(alt,ep)	(&(alt)->endpoint[ep])
 #define get_ep_desc(ep)		(ep)
@@ -384,6 +391,10 @@ static inline void *snd_compat_vmap(struct page **pages, unsigned int count, uns
 #define vmap snd_compat_vmap
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0) /* correct version? */
+#define EXPORT_NO_SYMBOLS
+#endif
+
 #include "amagic.h"
 
 #endif /* __SOUND_LOCAL_DRIVER_H */
